;;; cells.lisp
;;;
;;; Copyright (C) 2008-2009 Alessio Stalla
;;;
;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License
;;; as published by the Free Software Foundation; either version 2
;;; of the License, or (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
;;;
;;; As a special exception, the copyright holders of this library give you
;;; permission to link this library with independent modules to produce an
;;; executable, regardless of the license terms of these independent
;;; modules, and to copy and distribute the resulting executable under
;;; terms of your choice, provided that you also meet, for each linked
;;; independent module, the terms and conditions of the license of that
;;; module.  An independent module is a module which is not derived from
;;; or based on this library.  If you modify this library, you may extend
;;; this exception to your version of the library, but you are not
;;; obligated to do so.  If you do not wish to do so, delete this
;;; exception statement from your version.

(in-package :snow)

(defmodel cell-expression ()
  ((expression :initarg :expression :accessor c-value
	       :initform (error "expression is mandatory")
	       :cell t)))

(defun c-expr (&optional initial-value)
  (make-instance 'cell-expression :expression (c-in initial-value)))


(defobserver c-value ((x cell-expression) new-value)
  (format t "nv ~A ~A~%" x new-value))

;;Cellular slot Binding
(defmodel cell-data-binding (data-binding cells::model-object)
  ((expression :initarg :expression :reader binding-expression
	       :initform (error "expression is mandatory")
	       :cell t)
   (writer :initarg writer :accessor binding-writer :initform nil :cell nil)
   (model :accessor binding-model :initform nil :cell nil)))

(defmethod initialize-instance :after ((obj cell-data-binding) &rest args)
  (declare (ignore args))
  (setf (binding-model obj)
	(make-cells-value-model obj)))

(defobserver expression ((binding cell-data-binding) new-value)
  (bwhen (it (binding-model binding))
    (invoke "valueChanged" it new-value)))

(defun make-cell-data-binding (expression &optional writer)
  (check-type writer (or null function))
  (let ((instance
	 (make-instance 'cell-data-binding :expression expression)))
    (setf (binding-writer instance) writer)
    instance))

(defun make-slot-data-binding (object slot-accessor-name)
  (make-cell-data-binding
   (eval `(c? (,slot-accessor-name ,object)))
   (compile nil `(lambda (x)
		   (setf (,slot-accessor-name ,object) x)))))

(defmethod make-model ((binding cell-data-binding))
  (binding-model binding))
  
(defun make-cells-value-model (binding)
  (new "snow.binding.AccessorBinding"
       binding
       #'binding-expression
       (lambda (value place)
	 (declare (ignore place))
	 (bwhen (it (binding-writer binding))
	   (funcall it value)))))